import { helpers, RouterContext } from "https://deno.land/x/oak@v7.3.0/mod.ts";
import { compareMunicipalRibbon } from "./image.ts";
import { decode, Image } from "https://deno.land/x/jpegts@1.1/mod.ts";
import { parse as parseCsv } from "https://deno.land/std@0.97.0/encoding/csv.ts";

export interface Ribbon {
  acceptance: string;
  author?: string;
  cancellation?: string;
  cancellationReason?: string;
  description: string;
  figure?: string[];
  historicalJustification: string;
  img: string;
  municipalityName: string;
  municipality: string;
}

interface GetParams {
  from?: number;
  to?: number;
}

interface Category {
  name: string;
  hitrate: number;
}

export const getAllRibbons = (ctx: RouterContext) => {
  const params: GetParams = helpers.getQuery(ctx, { mergeParams: true });

  if (!(params.from === undefined || params.to === undefined)) {
    const from = Number(params.from);
    const to = Number(params.to);

    if (from > to) {
      ctx.response.status = 400;
      ctx.response.body = `'from' ${from} is bigger than 'to' ${to}!`;
      return;
    }

    ctx.response.body = ctx.state.slice(from, to) as Ribbon[];
  } else {
    ctx.response.body = ctx.state as Ribbon[];
  }

  ctx.response.status = 200;
};

export const getRibbon = ({ params, response, state }: RouterContext) => {
  if (params?.municipality === undefined) {
    response.status = 400;
    return;
  }
  const municipality = params?.municipality;
  const ribbon = state.find(
    (ribbon: Ribbon) => ribbon.municipality.endsWith(`/${municipality}`),
  );

  if (!ribbon) {
    response.status = 404;
    return;
  }

  response.status = 200;
  response.body = ribbon;
};

export const getRibbonCategories = async (
  { params, response, state }: RouterContext,
) => {
  if (params?.municipality === undefined) {
    response.status = 400;
    return;
  }

  const municipality = params?.municipality;
  const ribbon: Ribbon = state.find(
    (ribbon: Ribbon) => ribbon.municipality.endsWith(`/${municipality}`),
  );

  if (!ribbon) {
    response.status = 404;
    return;
  }

  if (!ribbon.figure) {
    response.status = 404;
    response.body = [];
    return;
  }

  const categories: Category[] = [];
  const content = await parseCsv(
    await Deno.readTextFile("categories.csv"),
  ) as string[][];
  for (let i = 1; i < content.length; i++) {
    const col = content[i];
    for (let j = 1; j < col.length; j++) {
      const figure = col[j];
      if (
        ribbon.figure.find((obj) => obj == figure)
      ) {
        switch (j) {
          case 1:
            findOrAddCategory(categories, "animals");
            break;
          case 2:
            findOrAddCategory(categories, "religionAndEpics");
            break;
          case 3:
            findOrAddCategory(categories, "itemsAndWeapons");
            break;
          case 4:
            findOrAddCategory(categories, "structuresAndAgriculture");
            break;
          case 5:
            findOrAddCategory(categories, "natureAndPlants");
            break;
          case 6:
            findOrAddCategory(categories, "water");
            break;
          default:
            throw new Error(`Unknown Category ${j}`);
        }
      }
    }
  }

  let totalHits = 0;

  categories.forEach((obj) => totalHits += obj.hitrate);

  for (const category of categories) {
    category.hitrate /= totalHits;
  }

  categories.sort((a, b) => a.hitrate - b.hitrate);

  response.status = 200;
  response.body = categories;
};

const findOrAddCategory = (categories: Category[], name: string) => {
  const find = categories.find((obj) => obj.name == name);
  if (find) {
    find.hitrate++;
  } else {
    categories.push({ name: name, hitrate: 1 });
  }
};

export const compareRibbon = async (
  { params, request, response, state }: RouterContext,
) => {
  if (params?.municipality === undefined) {
    response.status = 400;
    return;
  }
  const municipality = params?.municipality;
  const ribbon: Ribbon = state.find(
    (ribbon: Ribbon) => ribbon.municipality.endsWith(`/${municipality}`),
  );

  if (ribbon === undefined) {
    response.status = 404;
    response.body = "Couldn't find municipality " + municipality;
    return;
  }

  const { value } = request.body({ type: "form-data" });
  const { files } = await value.read();
  if (!files) {
    response.status = 400;
    return;
  }
  const filename = files[0].filename;
  if (!filename) {
    response.status = 400;
    return;
  }

  let img: Image | undefined = undefined;
  try {
    img = decode(
      await Deno.readFile(filename),
    );
  } catch (e) {
    response.status = 500;
    response.body = "" + e;
    return;
  }

  let matchPercentage = 0;
  try {
    matchPercentage = await compareMunicipalRibbon(
      ribbon.img,
      img,
    );
  } catch (e) {
    response.status = 500;
    response.body = "" + e;
    return;
  }

  response.body = { matchPercentage: matchPercentage };
};
